import glob
import os
import re
from urllib.parse import urlparse

from oelint_adv.cls_item import Variable
from oelint_adv.const_vars import get_known_mirrors


def get_files(stash, _file, pattern):
    """Get files matching SRC_URI entries

    Arguments:
        stash {oelint_adv.cls_stash.Stash} -- current stash
        _file {str} -- Full path to filename
        pattern {str} -- glob pattern to apply

    Returns:
        list -- list of files matching pattern
    """
    res = []
    src_uris = stash.GetItemsFor(filename=_file, classifier=Variable.CLASSIFIER,
                                 attribute=Variable.ATTR_VAR, attributeValue="SRC_URI")
    files_paths = list(
        set(["{}/*/{}".format(os.path.dirname(x.Origin), pattern) for x in src_uris]))
    for item in src_uris:
        files_paths += list(set(["{}/*/{}".format(os.path.dirname(x.Origin), pattern)
                                 for x in stash.GetItemsFor(filename=item.Origin)]))
    for item in files_paths:
        res += glob.glob(item)
    return list(set(res))


def find_local_or_in_layer(name, localdir):
    """Find file in local dir or in layer

    Arguments:
        name {str} -- filename
        localdir {str} -- path to local dir

    Returns:
        str -- path to found file or None
    """
    if os.path.exists(os.path.join(localdir, name)):
        return os.path.join(localdir, name)
    _curdir = localdir
    while os.path.isdir(_curdir):
        if _curdir == "/":
            break
        _curdir = os.path.dirname(_curdir)
        if os.path.exists(os.path.join(_curdir, "conf/layer.conf")):
            if os.path.exists(os.path.join(_curdir, name)):
                return os.path.join(_curdir, name)
            else:
                break
    return None


def _replace_with_known_mirrors(_in):
    """
    Replace the known mirror configuration items
    """
    for k, v in get_known_mirrors().items():
        _in = _in.replace(k, v)
    return _in


def get_scr_components(string):
    """Return SRC_URI components

    Arguments:
        string {str} -- raw string

    Returns:
        dict -- scheme: protocol used, src: source URI, options: parsed options
    """
    _url = urlparse(_replace_with_known_mirrors(string))
    _scheme = _url.scheme
    _tmp = _url.netloc
    if _url.path:
        _tmp += "/" + _url.path.lstrip("/")
    _path = _tmp.split(";")[0]
    _options = _tmp.split(";")[1:]
    _parsed_opt = {x.split("=")[0]: x.split("=")[1] for x in _options if "=" in x}
    return {"scheme": _scheme, "src": _path, "options": _parsed_opt}


def safe_linesplit(string):
    """Split line in a safe manner 

    Arguments:
        string {str} -- raw input

    Returns:
        list -- safely split input
    """
    return re.split(r"\s|\t|\x1b", string)

def guess_recipe_name(_file):
    """Get the recipe name from filename

    Arguments:
        _file {str} -- filename

    Returns:
        str -- recipe name
    """
    _name, _ = os.path.splitext(os.path.basename(_file))
    return _name.split("_")[0]

def guess_recipe_version(_file):
    """Get recipe version from filename

    Arguments:
        _file {str} -- filename

    Returns:
        str -- recipe version
    """
    _name, _ = os.path.splitext(os.path.basename(_file))
    return _name.split("_")[-1]

def expand_term(stash, _file, value):
    """Expand a variable (replacing all variables by known content)

    Arguments:
        stash {oelint_adv.cls_stash.Stash} -- current stash
        _file {str} -- Full path to file
        value {str} -- Variable value to expand

    Returns:
        str -- expanded value
    """
    pattern = r"\$\{(.+?)\}"
    res = str(value)
    for m in re.finditer(pattern, value):
        _comp = stash.GetItemsFor(filename=_file, classifier=Variable.CLASSIFIER,
                                  attribute=Variable.ATTR_VAR, attributeValue=m.group(1))
        if any(_comp):
            res = res.replace(m.group(0), expand_term(stash, _file, _comp[0].VarValueStripped))
        elif m.group(1) in ["PN", "BPN"]:
            res = res.replace(m.group(0), guess_recipe_name(_file))
        elif m.group(1) in ["PV"]:
            res = res.replace(m.group(0), guess_recipe_version(_file))
    return res

def get_valid_package_names(stash, _file):
    """Get known valid names for packages

    Arguments:
        stash {oelint_adv.cls_stash.Stash} -- current stash
        _file {str} -- Full path to file

    Returns:
        list -- list of valid package names
    """
    res = set()
    _comp = stash.GetItemsFor(filename=_file, classifier=Variable.CLASSIFIER,
                              attribute=Variable.ATTR_VAR, attributeValue="PACKAGES")
    _recipe_name = guess_recipe_name(_file)
    res.add(_recipe_name)
    for item in _comp:
        for pkg in [x for x in safe_linesplit(item.VarValueStripped) if x]:
            _pkg = pkg.replace("${PN}", _recipe_name)
            res.add(_pkg)
    return res

def get_valid_named_resources(stash, _file):
    """Get list of valid SRCREV resource names

    Arguments:
        stash {oelint_adv.cls_stash.Stash} -- current stash
        _file {str} -- Full path to file

    Returns:
        list -- list of valid SRCREV resource names
    """
    res = set()
    _comp = stash.GetItemsFor(filename=_file, classifier=Variable.CLASSIFIER,
                              attribute=Variable.ATTR_VAR, attributeValue="SRC_URI")
    _recipe_name = guess_recipe_name(_file)
    res.add(_recipe_name)
    for item in _comp:
        for name in [x for x in safe_linesplit(item.VarValueStripped) if x]:
            _url = get_scr_components(name)
            if "name" in _url["options"]:
                res.add(_url["options"]["name"].replace("${PN}", _recipe_name))
    return res



from oelint_adv.cls_rule import Rule
from oelint_adv.cls_item import PythonBlock, Function


class NoCommentsTrailing(Rule):
    def __init__(self):
        super().__init__(id="oelint.comments.notrailing",
                         severity="error",
                         message="Comments shall be put on seperate lines")

    def check(self, _file, stash):
        res = []
        items = stash.GetItemsFor(filename=_file)
        for i in items:
            if isinstance(i, PythonBlock) or isinstance(i, Function):
                continue
            if i.Raw:
                lines = i.Raw.split("\n")
                for line in lines:
                    _line = line.strip()
                    if "#" in _line and _line.find("#") > 0:
                        res += self.finding(i.Origin, i.InFileLine + lines.index(line))
        return res



import collections
import os
import re

from oelint_adv.cls_item import Comment
from oelint_adv.cls_item import Function
from oelint_adv.cls_item import Include
from oelint_adv.cls_item import Item
from oelint_adv.cls_item import PythonBlock
from oelint_adv.cls_item import TaskAdd
from oelint_adv.cls_item import TaskAssignment
from oelint_adv.cls_item import Variable
from oelint_adv.helper_files import find_local_or_in_layer, expand_term


def get_full_scope(_string, offset, _sstart, _send):
    scopelevel = 0
    pos = 0
    for c in _string[offset:]:
        if c == _sstart:
            scopelevel += 1
        elif c == _send:
            scopelevel -= 1
        pos += 1
        if scopelevel < 0:
            break
    return _string[:pos+offset]


def prepare_lines_subparser(_iter, lineOffset, num, line, raw_line=None):
    __func_start_regexp__ = r".*(((?P<py>python)|(?P<fr>fakeroot))\s*)*(?P<func>[\w\.\-\+\{\}\$]+)?\s*\(\s*\)\s*\{"
    res = []
    raw_line = raw_line or line
    if raw_line.find("\\\n") != -1:
        _, line = _iter.__next__()
        while line.find("\\\n") != -1:
            raw_line += line
            _, line = _iter.__next__()
        raw_line += line
    elif re.match(__func_start_regexp__, raw_line):
        _, line = _iter.__next__()
        stopiter = False
        while line.strip() != "}" and not stopiter:
            raw_line += line
            try:
                _, line = _iter.__next__()
            except StopIteration:
                stopiter = True
        if line.strip() == "}":
            raw_line += line
    elif raw_line.strip().startswith("def "):
        stopiter = False
        while not stopiter:
            try:
                _, line = _iter.__next__()
            except StopIteration:
                stopiter = True
            if line.startswith("def ") or line.startswith("#"):
                raw_line = line
                res += prepare_lines_subparser(_iter,
                                               lineOffset, num, line, raw_line=raw_line)
                break
            if not line.startswith(" ") and not line.startswith("\t"):
                break
            raw_line += line

    while raw_line.find("${@") != -1:
        _inline_block = raw_line.find("${@")
        repl = get_full_scope(raw_line[_inline_block:], len("${@"), "{", "}")
        raw_line = raw_line.replace(repl, "!!!inlineblock!!!")
    res.append({"line": num + 1 + lineOffset, "raw": raw_line,
                "cnt": raw_line.replace("\n", "").replace("\\", chr(0x1b))})
    return res


def prepare_lines(_file, lineOffset=0):
    try:
        prep_lines = []
        with open(_file) as i:
            _iter = enumerate(i.readlines())
            for num, line in _iter:
                prep_lines += prepare_lines_subparser(
                    _iter, lineOffset, num, line)
    except FileNotFoundError:
        pass
    return prep_lines


def get_items(stash, _file, lineOffset=0):
    res = []
    __regex_var = r"^(?P<varname>([A-Z0-9a-z_-]|\$|\{|\})+?)(\[(?P<ident>(\w|-)+)\])*(?P<varop>(\s|\t)*(\+|\?|\:|\.)*=(\+)*(\s|\t)*)(?P<varval>.*)"
    __regex_func = r"^((?P<py>python)\s+|(?P<fr>fakeroot\s+))*(?P<func>[\w\.\-\+\{\}\$]+)?\s*\(\s*\)\s*\{(?P<funcbody>.*)\s*\}"
    __regex_inherit = r"^.*?inherit(\s+|\t+)(?P<inhname>.+)"
    __regex_comments = r"^(\s|\t)*#+\s*(?P<body>.*)"
    __regex_python = r"^(\s*|\t*)def(\s+|\t+)(?P<funcname>[a-z0-9_]+)(\s*|\t*)\(.*\)\:"
    __regex_include = r"^(\s*|\t*)(?P<statement>include|require)(\s+|\t+)(?P<incname>[A-za-z0-9\-\./\$\{\}]+)"
    __regex_addtask = r"^(\s*|\t*)addtask\s+(?P<func>\w+)\s*((before\s*(?P<before>((.*(?=after))|(.*))))|(after\s*(?P<after>((.*(?=before))|(.*)))))*"
    __regex_taskass = r"^(\s*|\t*)(?P<func>[a-z0-9_-]+)\[(?P<ident>\w+)\](\s+|\t+)=(\s+|\t+)(?P<varval>.*)"

    _order = collections.OrderedDict([
        ("comment", __regex_comments),
        ("func", __regex_func),
        ("inherit", __regex_inherit),
        ("python", __regex_python),
        ("include", __regex_include),
        ("addtask", __regex_addtask),
        ("taskassign", __regex_taskass),
        ("vars", __regex_var)
    ])

    includeOffset = 0

    for line in prepare_lines(_file, lineOffset):
        good = False
        for k, v in _order.items():
            m = re.match(v, line["cnt"], re.MULTILINE)
            if m:
                if k == "python":
                    res.append(PythonBlock(
                        _file, line["line"] + includeOffset, line["line"] - lineOffset, line["raw"], m.group("funcname")))
                    good = True
                    break
                elif k == "vars":
                    res.append(Variable(
                        _file, line["line"] + includeOffset, line["line"] -
                        lineOffset, line["raw"], m.group(
                            "varname"), m.group("varval"),
                        m.group("varop"), m.group("ident")))
                    good = True
                    break
                elif k == "func":
                    res.append(Function(
                        _file, line["line"] + includeOffset, line["line"] -
                        lineOffset, line["raw"],
                        m.group("func"), m.group("funcbody"),
                        m.group("py"), m.group("fr")))
                    good = True
                    break
                elif k == "comment":
                    res.append(
                        Comment(_file, line["line"] + includeOffset, line["line"] - lineOffset, line["raw"]))
                    good = True
                    break
                elif k == "inherit":
                    res.append(Variable(
                        _file, line["line"] + includeOffset, line["line"] -
                        lineOffset, line["raw"], "inherit", m.group("inhname"),
                        "", ""))
                    good = True
                    break
                elif k == "taskassign":
                    res.append(TaskAssignment(_file, line["line"] + includeOffset, line["line"] - lineOffset, line["raw"], m.group(
                        "func"), m.group("ident"), m.group("varval")))
                    good = True
                    break
                elif k == "addtask":
                    # treat the following as variables
                    if any([m.group("func").startswith(x) for x in ['pkg_preinst', 'pkg_postinst', 'pkg_prerm', 'pkg_postrm']]):
                        continue
                    _g = m.groupdict()
                    if "before" in _g.keys():
                        _b = _g["before"]
                    else:
                        _b = ""
                    if "after" in _g.keys():
                        _a = _g["after"]
                    else:
                        _a = ""
                    res.append(TaskAdd(
                        _file, line["line"] + includeOffset, line["line"] - lineOffset, line["raw"], m.group("func"), _b, _a))
                    break
                elif k == "include":
                    _path = find_local_or_in_layer(
                        expand_term(stash, _file, m.group("incname")), os.path.dirname(_file))
                    if _path:
                        tmp = stash.AddFile(
                            _path, lineOffset=line["line"], forcedLink=_file)
                        if any(tmp):
                            includeOffset += max([x.InFileLine for x in tmp])
                    res.append(Include(
                        _file, line["line"], line["line"] - lineOffset, line["raw"], m.group("incname"), m.group("statement")))
                    good = True
                    break
        if not good:
            res.append(
                Item(_file, line["line"], line["line"] - lineOffset, line["raw"]))
    return res
